#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include <string.h>
#include "pmandel.h"

#define  MASTER  0
#define  TAG 17

#ifndef NROWS
#define NROWS 20
#endif

int         np, myid, nrows;
MPI_Datatype MPI_Param, MPI_Complex;
MPE_Color *colorArray;

/* To allow a fair comparison between several runs determine zoom areas in advannce */
#define NUM_ZOOMS 3

/* Logging on master */
#define MPE_START_BCAST_PARAM  1701
#define MPE_STOP_BCAST_PARAM   1702
#define MPE_START_SEND_JOB    1703
#define MPE_STOP_SEND_JOB     1704

/* Logging on slaves */
#define MPE_START_CALC        1707
#define MPE_STOP_CALC         1708
#define MPE_START_COMMUNICATE 1709
#define MPE_STOP_COMMUNICATE  1710

Complex_t cStarts[NUM_ZOOMS] = {{ -6.025000e-01, -6.600000e-01},
                                { -5.092736e-01, -6.076833e-01},
                                { -5.074147e-01, -6.022850e-01}};

Complex_t cSteps[NUM_ZOOMS] = {{4.180556e-04, 3.583333e-04},
                               {1.602546e-05, 1.373611e-05},
                               {1.175201e-06, 1.007315e-06}};

void print_parameters(Parameter_t *Param)
{
	printf("New parameters:\n");
	printf("\tcStart = (%e, %e)\n", Param->cStart.real, Param->cStart.imag);
	printf("\tcStep  = (%e, %e)\n", Param->cStep.real,  Param->cStep.imag);
	fflush(stdout);
}

int DragZoom(Parameter_t *Param, Winspecs *winspecs)
{
	int xLeft, yTop, xRight, yBottom, i, button;

	MPE_Get_drag_region(winspecs->graph, 1, MPE_DRAG_SQUARE,
	                    &xLeft, &yTop, &xRight, &yBottom);
	Param->cStart.real += min(xLeft, xRight) * Param->cStep.real;
	Param->cStart.imag += min(yTop, yBottom) * Param->cStep.imag;
	Param->cStep.real *= (double) + abs(xRight - xLeft) / Param->PicWidth;
	Param->cStep.imag *= (double) + abs(yBottom - yTop) / Param->PicHeight;
	if (abs(xLeft - xRight) < 5)
		Param->CurrentLine = -1;
	return 1;
}

/* Set Param to next zoom area */
int NextZoom(Parameter_t *Param, Winspecs *winspecs)
{
	static int n = 0;
	int xLeft, yTop, xRight, yBottom, i, button;

	if (n < NUM_ZOOMS) {
		Param->cStart = cStarts[n];
		Param->cStep = cSteps[n];
	} else
		Param->CurrentLine = -1;
	n++;
	return 1;
}

void InitParams(Parameter_t *Param, Winspecs *winspecs)
{
	Param->CurrentLine = 0;
	Param->nRows       = nrows;
	Param->PicWidth    = winspecs->width;
	Param->PicHeight   = winspecs->height;
	Param->cStart.real = -2.3;
	Param->cStart.imag = -1.5;
	Param->cStep.real  = 3.5 / Param->PicWidth;
	Param->cStep.imag  = 3.0 / Param->PicHeight;
}

int doit(Parameter_t *Param, Winspecs *winspecs)
{
	static MPE_Point  *ImageData = NULL;

	if (ImageData == NULL)
		ImageData = (MPE_Point*) malloc((Param->nRows)
		                                * Param->PicWidth * sizeof(MPE_Point));

	MPE_Log_event(MPE_START_CALC, 0, "start calculating");
	calculate(ImageData, Param);
	MPE_Log_event(MPE_STOP_CALC, 0, "stop calculating");

	MPE_Draw_points(winspecs->graph, ImageData,
	                Param->nRows*Param->PicWidth);
	MPE_Update(winspecs->graph);
}

int Master(Parameter_t *Param, Winspecs *winspecs)
{
	int rank, dummy;
	MPI_Status status;
	struct timeval tv1, tv2;
	float usec;

	gettimeofday(&tv1, NULL);
	Param->CurrentLine = 0;

	/* Send first jobs */
	for (rank = 1; rank < np; rank++) {
		MPE_Log_event(MPE_START_SEND_JOB, 0, "start sending job");
		MPI_Send(&Param->CurrentLine, 1, MPI_INT, rank, WORK_TAG, MPI_COMM_WORLD);
		MPE_Log_event(MPE_STOP_SEND_JOB, 0, "stop sending job");

		Param->CurrentLine += Param->nRows;
	}

	/* Send next jobs in first-come-first-serve manner */
	while (Param->CurrentLine <= Param->PicHeight) {
		MPI_Recv(&dummy, 0, MPI_INT, MPI_ANY_SOURCE, DONE_TAG, MPI_COMM_WORLD, &status);

		MPE_Log_event(MPE_START_SEND_JOB, 0, "start sending job");
		MPI_Send(&Param->CurrentLine, 1, MPI_INT, status.MPI_SOURCE, WORK_TAG, MPI_COMM_WORLD);
		MPE_Log_event(MPE_STOP_SEND_JOB, 0, "stop sending job");

		Param->CurrentLine += Param->nRows;
	}

	/* Wait for outstanding jobs to finish */
	for (rank = 1; rank < np; rank++)
		MPI_Recv(&dummy, 0, MPI_INT, rank, DONE_TAG, MPI_COMM_WORLD, &status);

	gettimeofday(&tv2, NULL);
	usec = (tv2.tv_sec - tv1.tv_sec) * 1e6 + (tv2.tv_usec - tv1.tv_usec);
	printf("elapsed time: %.2f seconds\n", usec / 1e6);

	/* Tell all the slaves to exit calculation loop */
	dummy = -1;
	for (rank = 1; rank < np; rank++) {
		// MPE_Log_event(MPE_START_SEND_JOB, 0, "start sending job");
		MPI_Send(&dummy, 1, MPI_INT, rank, WORK_TAG, MPI_COMM_WORLD);
		// MPE_Log_event(MPE_STOP_SEND_JOB, 0, "stop sending job");
	}
	return 0;
}

int Slave(Parameter_t *Param, Winspecs *winspecs)
{
	MPI_Status status;
	int dummy;

	do {
		MPE_Log_event(MPE_START_COMMUNICATE, 0, "start communication");
		MPI_Recv(&Param->CurrentLine, 1, MPI_INT, 0, WORK_TAG, MPI_COMM_WORLD, &status);
		MPE_Log_event(MPE_STOP_COMMUNICATE, 0, "stop communication");

		if (Param->CurrentLine != -1) {
			doit(Param, winspecs);

			MPE_Log_event(MPE_START_COMMUNICATE, 0, "start communication");
			MPI_Send(&dummy, 0, MPI_INT, 0, DONE_TAG, MPI_COMM_WORLD); /* say I'm ready */
			MPE_Log_event(MPE_STOP_COMMUNICATE, 0, "stop communication");
		}
	} while (Param->CurrentLine != -1);
	return 0;
}

void InitDataTypes(void)
{
	Parameter_t aparam;
	MPI_Datatype Type[2];
	int Block[2];
	MPI_Aint Disp[2];

	Type[0] = MPI_DOUBLE;
	Block[0] = 2;
	Disp[0] = 0;
	MPI_Type_struct(1, Block, Disp, Type, &MPI_Complex);
	MPI_Type_commit(&MPI_Complex);

	Type[0] = MPI_INT; Type[1] = MPI_Complex;
	Block[0] = 4;      Block[1] = 2;
	MPI_Address(&aparam.CurrentLine, &Disp[0]);
	MPI_Address(&aparam.cStart, &Disp[1]);
	Disp[1] -= Disp[0];
	Disp[0] = 0;
	MPI_Type_struct(2, Block, Disp, Type, &MPI_Param);
	MPI_Type_commit(&MPI_Param);
}


main(int  argc, char **argv)
{
	int   myWindowOpened, allWindowsOpened;
	Parameter_t     Param;
	Winspecs  winspecs;
	MPI_Status  status;
	int  i, j, k;
	char  hostname[80];
	char            logfilename[80];

	MPI_Init(&argc, &argv);

	MPE_Init_log();
	MPE_Describe_state(MPE_START_BCAST_PARAM, MPE_STOP_BCAST_PARAM, "Send Parameter", "red");
	MPE_Describe_state(MPE_START_SEND_JOB, MPE_STOP_SEND_JOB, "Send Job", "blue");
	MPE_Describe_state(MPE_START_CALC, MPE_STOP_CALC, "Calculate Job", "green");
	MPE_Describe_state(MPE_START_COMMUNICATE, MPE_STOP_COMMUNICATE, "Communicate", "brown");

	MPI_Comm_rank(MPI_COMM_WORLD, &myid);
	MPI_Comm_size(MPI_COMM_WORLD, &np);
	gethostname(hostname, 80);
	printf("Hi from Task %d on %s\n", myid, hostname);
	fflush(stdout);

	if (!myid) {
		if (argc > 1)
			nrows = atoi(argv[1]);
		else
			nrows = NROWS;
		printf("Using %d lines per job\n", nrows); fflush(stdout);
	}

	MPI_Bcast(&nrows, 1, MPI_INT, 0, MPI_COMM_WORLD);

	myWindowOpened = InitGraphics(&winspecs);
	MPI_Allreduce(&myWindowOpened, &allWindowsOpened, 1, MPI_INT, MPI_LAND,
	              MPI_COMM_WORLD);

	if (allWindowsOpened) {

		MakeColors(&winspecs);
		InitDataTypes();
		InitParams(&Param, &winspecs);

		do {
			if (!myid) {
				Master(&Param, &winspecs);
#ifdef MANUAL_ZOOM
				DragZoom(&Param, &winspecs);
#else
				NextZoom(&Param, &winspecs);
#endif
			} else {
				Slave(&Param, &winspecs);
			}
			/* distribute new parameters */
			MPE_Log_event(MPE_START_BCAST_PARAM, 0, "start sending/receiving parameters");
			MPI_Bcast(&Param, 1, MPI_Param, 0, MPI_COMM_WORLD);
			MPE_Log_event(MPE_STOP_BCAST_PARAM, 0, "stop sending/receiving parameters");
		} while (Param.CurrentLine != -1);

		printf("Task %d on %s ready\n", myid, hostname);
		snprintf(logfilename, 80, "%s%d", "pmandellog", nrows);
	
		MPI_Barrier(MPI_COMM_WORLD);
		MPE_Close_graphics(&(winspecs.graph));
	} else {
		if (!myid) {
			fprintf(stderr, "One or more processes could not connect\n");
			fprintf(stderr, "to the display.  Exiting.\n\n");
		}
		if (myWindowOpened)
			MPE_Close_graphics(&(winspecs.graph));
	}

	MPE_Finish_log("pmandel_state.log");
	MPI_Finalize();
}
